#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <time.h>

#include "turtle.h"
#include "kdtree.h"

#define ROTX(X,Y,THETA) ((X)*cos((THETA))-(Y)*sin((THETA)))
#define ROTY(X,Y,THETA) ((X)*sin((THETA))+(Y)*cos((THETA)))

// Draws the ship at (tx, ty) rotated by theta radians
void draw_ship(double tx, double ty, double theta) {
    turtle_set_pen_color(255, 255, 255);
    turtle_pen_up();
    turtle_goto(ROTX(-12,-12,theta)+tx, ROTY(-12,-12,theta)+ty);
    turtle_pen_down();
    turtle_goto(ROTX(18,0,theta)+tx, ROTY(18,0,theta)+ty);
    turtle_goto(ROTX(-12,12,theta)+tx, ROTY(-12,12,theta)+ty);
    turtle_goto(ROTX(-12,-12,theta)+tx, ROTY(-12,-12,theta)+ty);
}

// Draws a debris field with a given color
void draw_debris_field(debris_t debris[], size_t debris_len, int r, int g, int b) {
    const int sz = 2;
    turtle_pen_up();
    turtle_set_pen_color(r, g, b);
    turtle_set_fill_color(r, g, b);
    for (size_t i = 0; i < debris_len; i++) {
        turtle_goto(debris[i].pos.x-sz, debris[i].pos.y-sz);
        turtle_pen_down();
        turtle_begin_fill();
        turtle_goto(debris[i].pos.x-sz, debris[i].pos.y+sz);
        turtle_goto(debris[i].pos.x+sz, debris[i].pos.y+sz);
        turtle_goto(debris[i].pos.x+sz, debris[i].pos.y-sz);
        turtle_goto(debris[i].pos.x-sz, debris[i].pos.y-sz);
        turtle_end_fill();
        turtle_pen_up();
    }
}

// Draws a blue circle centered on a given point with a given radius
void draw_circle(point_t c, float radius) {
    turtle_set_pen_color(0, 0, 255);
    bool penup = true;
    turtle_pen_up();
    for (int i=0; i < 360; i+=4) {
        float rad = (i+(0*2))*3.14159f/180;
        turtle_goto((int)(c.x + cos(rad)*radius), (int)(c.y + sin(rad)*radius));
        if (penup) {
            turtle_pen_down();
            penup = false;
        } else {
            turtle_pen_up();
            penup = true;
        }
    }
}

// Draws a rectangle as specified by a bbox_t struct
void draw_box(bbox_t box) {
    turtle_pen_up();
    turtle_goto(box.minx, box.maxy);
    turtle_pen_down();
    turtle_goto(box.minx, box.miny);
    turtle_goto(box.maxx, box.miny);
    turtle_goto(box.maxx, box.maxy);
    turtle_goto(box.minx, box.maxy);
}

// Draws a dashed vertical line
void draw_dashed_vertical_line(float x, float miny, float maxy) {
    turtle_pen_up();
    turtle_goto_real(x, miny);
    turtle_set_heading(90);
    float y = miny;
    while (y < maxy-2) {
        turtle_pen_down();
        turtle_forward(2);
        turtle_pen_up();
        turtle_forward(3);
        y += 5;
    }
}

// Draws a dashed horizontal line
void draw_dashed_horizontal_line(float y, float minx, float maxx) {
    turtle_pen_up();
    turtle_goto_real(minx, y);
    turtle_set_heading(0);
    float x = minx;
    while (x < maxx-2) {
        turtle_pen_down();
        turtle_forward(2);
        turtle_pen_up();
        turtle_forward(3);
        x += 5;
    }
}

// Draws a dashed red line to show the current node split, then recurse on the
// left and right side
void draw_kdtree_node(kdtree_node_t *node, bbox_t bounds, size_t depth) {
    if (node == NULL) {
        return;
    }
    if (depth % 2 == 0) {

        // draw vertical line
        float x = node->debris.pos.x;
        turtle_set_pen_color(255, 0, 0);
        draw_dashed_vertical_line(x, bounds.miny, bounds.maxy);

        // recurse on left and right side
        bbox_t new_bounds = bounds;
        new_bounds.maxx = x;
        draw_kdtree_node(node->left, new_bounds, depth+1);
        new_bounds = bounds;
        new_bounds.minx = x;
        draw_kdtree_node(node->right, new_bounds, depth+1);

    } else {

        // draw horizontal line
        float y = node->debris.pos.y;
        turtle_set_pen_color(255, 0, 0);
        draw_dashed_horizontal_line(y, bounds.minx, bounds.maxx);

        // recurse on left and right side
        bbox_t new_bounds = bounds;
        new_bounds.maxy = y;
        draw_kdtree_node(node->left, new_bounds, depth+1);
        new_bounds = bounds;
        new_bounds.miny = y;
        draw_kdtree_node(node->right, new_bounds, depth+1);
    }
}

// Draws kD-tree structure (useful for debugging)
void draw_kdtree(kdtree_t *tree) {
    bbox_t bounds = {.minx = -300.f, .maxx = 300.f,
                     .miny = -300.f, .maxy = 300.f};
    draw_kdtree_node(tree->root, bounds, 0);
}

// Create a random debris field
void create_random_debris_tree(kdtree_t *tree) {
    size_t debris_count = 128;
    debris_t debris_array_1[debris_count];
    for (int i = 0; i < debris_count; i++) {
        debris_array_1[i].pos.x = 598 * (float) rand() / (float) RAND_MAX - 299;
        debris_array_1[i].pos.y = 598 * (float) rand() / (float) RAND_MAX - 299;
        debris_array_1[i].active = true;
    }
    kdtree_build(tree, debris_array_1, debris_count);
}

int main() {

    // Initialize random number generator
    srand(time(NULL));

    // Initalize turtle graphics
    turtle_init(600, 600);

    // Draw a black background
    turtle_set_fill_color(0, 0, 0);
    turtle_goto(-300, -300);
    turtle_begin_fill();
    turtle_goto(300, -300);
    turtle_goto(300, 300);
    turtle_goto(-300, 300);
    turtle_goto(-300, -300);
    turtle_end_fill();

    // Create the kD-tree with the debris from the debris field
    kdtree_t tree;
    kdtree_init(&tree);
    create_random_debris_tree(&tree);

    // Draw the kD-tree structure (enable this to help with debugging)
    /*draw_kdtree(&tree);*/

    // Draw the debris field by doing a breadth-first traversal of the tree
    if (tree.size > 0) {
        debris_t* debris_array = (debris_t *)malloc(sizeof(debris_t) * tree.size);

        // Initialize queue
        kdtree_node_t** queue =
            (kdtree_node_t **)malloc(sizeof(kdtree_node_t*) * tree.size);
        size_t queue_back = 0;
        size_t queue_front = 0;

        // Push the root
        queue[queue_back++] = tree.root;
        int debris_array_idx = 0;

        // While the queue is not empty, continue the BFS
        while (queue_front != queue_back) {
            kdtree_node_t *cur_node = queue[queue_front++];
            if (cur_node->left) queue[queue_back++] = cur_node->left;
            if (cur_node->right) queue[queue_back++] = cur_node->right;
            debris_array[debris_array_idx++] = cur_node->debris;
        }

        // Draw debris and clean up memory
        draw_debris_field(debris_array, tree.size, 255, 255, 255);
        free(queue);
        free(debris_array);
    }

    // Draw the player's ship
    point_t ship_pos = {.x = 100.f, .y = 75.f};
    draw_ship(ship_pos.x, ship_pos.y, 3.14159f/8);

    // Box range searching -- first, draw the box (in green)
    bbox_t range = {.minx = -200,       .miny = -150,
                    .maxx = ship_pos.x, .maxy = ship_pos.y};
    turtle_set_pen_color(0, 255, 0);
    draw_box(range);

    // Search for points in the box
    dynarray_t box_pts;
    dynarray_init(&box_pts);
    kdtree_box_range_search(&tree, range, &box_pts);

    // Copy the points from the dynarray_t into an array and draw them in green
    size_t box_size = dynarray_size(&box_pts);
    if (box_size > 0) {
        debris_t *box_debris = malloc(sizeof(debris_t) * box_size);
        for (size_t i = 0; i < box_size; i++) {
            box_debris[i] = *(dynarray_pop(&box_pts));
        }
        draw_debris_field(box_debris, box_size, 0, 255, 0);
        free(box_debris);
    }

    // Circle range searching -- first, draw the circle (in blue)
    draw_circle(ship_pos, 80.f);

    // Search for points in the circle
    dynarray_t circ_pts;
    dynarray_init(&circ_pts);
    kdtree_circ_range_search(&tree, ship_pos, 80.f, &circ_pts);

    // Copy the points from the dynarray_t into an array and draw them in blue
    size_t circ_size = dynarray_size(&circ_pts);
    if (circ_size > 0) {
        debris_t *circ_debris = malloc(sizeof(debris_t) * circ_size);
        for (size_t i = 0; i < circ_size; i++) {
            circ_debris[i] = *(dynarray_pop(&circ_pts));
        }
        draw_debris_field(circ_debris, circ_size, 0, 0, 255);
        free(circ_debris);
    }

    // Save image and clean up memory
    turtle_save_bmp("rangefighter.bmp");
    turtle_cleanup();
    kdtree_free(&tree);
    dynarray_free(&circ_pts);
    dynarray_free(&box_pts);

    return EXIT_SUCCESS;
}
